
from logging import Logger
from logging import getLogger

from wx import OK
from wx import ICON_ERROR

from wx import MessageDialog

from umlextensions.ExtensionsPreferences import ExtensionsPreferences
from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.ExtensionDataTypes import Author
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName
from umlextensions.extensiontypes.ExtensionDataTypes import Version


class BaseExtension:
    """
    Contains common behavior and attributes for the various
    types of extensions
    """
    def __init__(self, extensionsFacade: IExtensionsFacade):

        self._extensionsFacade: IExtensionsFacade     = extensionsFacade
        self._preferences:      ExtensionsPreferences = ExtensionsPreferences()

        self._baseLogger: Logger = getLogger(__name__)

        self._name:         ExtensionName = ExtensionName('Implementor must provide the plugin name')
        self._author:       Author        = Author('Implementor must provide the plugin author')
        self._version:      Version       = Version('Implementor must provide the version')

        self._requireActiveFrame: bool = True
        """
        Extensions that require an active frame or frame(s) should set this value to `True`
        Some output extension may create their own frame or their own project and frame.  These should set this value to `False`
        Extensions should set the value they need in their constructor
        """
        self._requireSelection:   bool = True
        """
        Some Output extension may offer the option of exporting only selected shape;  Others may just export
        the entire project or the current frame

        Extensions should set the value they need in their constructor
        """
        self._autoSelectAll: bool = self._preferences.autoSelectAll
        """
        Some extensions may need to work with all the shapes on the UML frame.  Set this
        to `True` to select them all
        """

    @property
    def name(self) -> ExtensionName:
        """
        Implementations set the protected variable at class construction

        Returns:  The plugin name
        """
        return self._name

    @property
    def author(self) -> Author:
        """
        Implementations set the protected variable at class construction

        Returns:  The author's name
        """
        return self._author

    @property
    def version(self) -> Version:
        """
        Implementations set the protected variable at class construction

        Returns: The plugin version string
        """
        return self._version

    @classmethod
    def showNoUmlFrameDialog(cls):
        booBoo: MessageDialog = MessageDialog(parent=None, message='No UML frame', caption='Try Again!', style=OK | ICON_ERROR)
        booBoo.ShowModal()

    @classmethod
    def showNoSelectedUmlShapesDialog(cls):
        booBoo: MessageDialog = MessageDialog(parent=None, message='No selected UML shapes', caption='Try Again!', style=OK | ICON_ERROR)
        booBoo.ShowModal()




from typing import List
from typing import Dict
from typing import Union
from typing import NewType

from enum import Enum

from dataclasses import dataclass

from pyutmodelv2.PyutClass import PyutClass

VERSION: str = '3.0'

PyutClassName = NewType('PyutClassName', str)
ParentName    = NewType('ParentName',    str)
PropertyName  = NewType('PropertyName',  str)
ChildName     = NewType('ChildName',     str)

PropertyNames = NewType('PropertyNames', List[PropertyName])
PyutClasses   = NewType('PyutClasses',   Dict[PyutClassName, PyutClass])

PropertyMap   = NewType('PropertyMap',    Dict[PyutClassName, PropertyNames])
Children      = List[Union[PyutClassName, ChildName]]
Parents       = NewType('Parents',        Dict[ParentName,    Children])

AssociateName = PyutClassName

class AssociationType(Enum):

    ASSOCIATION = 'ASSOCIATION'
    AGGREGATION = 'AGGREGATION'
    COMPOSITION = 'COMPOSITION'
    INHERITANCE = 'INHERITANCE'
    INTERFACE   = 'INTERFACE'

@dataclass
class Associate:
    associateName:   AssociateName   = AssociateName('')
    associationType: AssociationType = AssociationType.ASSOCIATION


Associates = NewType('Associates', List[Associate])

#
# e.g.
#     @property
#     def pages(self) -> Pages:
#         return self._pages
# In the above "Pages" is the AssociateName and goes in the List for the method containing PyutClassName
#
# e.g.
#  self.pages: Pages = Pages({})
#
#  Pages is the AssociateName and the enclosing class for the __init__ method is the PyutClassName
#
#
Associations = NewType('Associations', Dict[PyutClassName, Associates])




from typing import cast

from logging import Logger
from logging import getLogger

from abc import ABC
from abc import abstractmethod

from wx import DirDialog
from wx import ID_CANCEL
from wx import DD_NEW_DIR_BUTTON

from umlextensions.ExtensionsTypes import FrameInformation
from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.BaseExtension import BaseExtension
from umlextensions.input.ImportDirectoryResponse import ImportDirectoryResponse

from umlextensions.input.InputFormat import InputFormat

from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionDescription
from umlextensions.extensiontypes.ExtensionDataTypes import FileSuffix
from umlextensions.extensiontypes.ExtensionDataTypes import FormatName

UNSPECIFIED_NAME:        FormatName           = FormatName('Unspecified Plugin Name')
UNSPECIFIED_FILE_SUFFIX: FileSuffix           = FileSuffix('*')
UNSPECIFIED_DESCRIPTION: ExtensionDescription = ExtensionDescription('Unspecified Extension Description')


class BaseInputExtension(BaseExtension, ABC):
    """
    Interface for extensions that can convert foreign structured
    data into UML Diagrams.  Examples include but are not limited to:

        * DTDs
        * Java code
        * Python code

    """
    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade)
        self._bieLogger: Logger = getLogger(__name__)

        self._inputFormat:      InputFormat      = InputFormat(formatName=UNSPECIFIED_NAME, fileSuffix=UNSPECIFIED_FILE_SUFFIX, description=UNSPECIFIED_DESCRIPTION)
        self._frameInformation: FrameInformation = cast(FrameInformation, None)

    @property
    def inputFormat(self) -> InputFormat:
        """
        Implementations set the protected variable at class construction

        Returns: The input format type; Plugins should return `None` if they do
        not support input operations
        """
        return self._inputFormat

    def askForImportDirectoryName(self) -> ImportDirectoryResponse:
        """
        Called by plugin to ask which directory must be imported

        Returns:  The appropriate response object;  The directory name is valid only if
        response.cancelled is True
        """
        # defaultPath: str = self._pluginAdapter.currentDirectory
        defaultPath: str = ''
        dirDialog: DirDialog = DirDialog(None,
                                         "Choose a directory to import",
                                         defaultPath=defaultPath,
                                         style=DD_NEW_DIR_BUTTON)

        response: ImportDirectoryResponse = ImportDirectoryResponse()
        if dirDialog.ShowModal() == ID_CANCEL:
            response.cancelled     = True
            response.directoryName = ''
        else:
            response.directoryName = dirDialog.GetPath()
            response.cancelled     = False
            # self._pluginAdapter.currentDirectory = response.directoryName    # TODO: Should plugin be doing this?  No

        dirDialog.Destroy()

        return response

    def executeImport(self):
        """
        Called by the extension manager to begin the import process.
        """
        # noinspection PyTypeChecker
        self._extensionsFacade.requestCurrentFrameInformation(callback=self._executeImport)   # type ignore

    def _executeImport(self, frameInformation: FrameInformation):
        """
        The callback necessary to start the import process;
        Args:
            frameInformation:
        """
        assert self.inputFormat is not None, 'Developer error. We cannot import w/o an import format'

        self._frameInformation = frameInformation

        if self._requireActiveFrame is True:
            if frameInformation.frameActive is False:
                self.showNoUmlFrameDialog()
                return
        if self.setImportOptions() is True:
            self.read()

    @abstractmethod
    def setImportOptions(self) -> bool:
        """
        Prepare for the import.
        Use this method to query the end-user for any additional import options

        Returns:
            if False, the import is cancelled
        """
        pass

    @abstractmethod
    def read(self) -> bool:
        """
        Read data from a file;  Presumably, the file was specified on the call
        to setImportOptions
        """
        pass

    def _composeWildCardSpecification(self) -> str:

        inputFormat: InputFormat = self.inputFormat

        # wildcard: str = inputFormat.name + " (*." + inputFormat.extension + ")|*." + inputFormat.description
        wildcard: str = (
            f'{inputFormat.formatName} '
            f' (*, {inputFormat.fileSuffix}) '
            f'|*.{inputFormat.fileSuffix}'
        )
        return wildcard



from typing import List

from umlshapes.frames.UmlFrame import UmlFrame

from umlshapes.shapes.UmlLineControlPoint import UmlLineControlPoint
from umlshapes.shapes.UmlLineControlPoint import UmlLineControlPointType

from umlextensions.tools.sugiyama.LayoutInterfaceLink import LayoutInterfaceLink
from umlextensions.tools.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode

#
# TODO: Figure out how to get that value here
# This is controlled by a UML Shapes preference
#
SUGIYAMA_CONTROL_POINT_SIZE: int = 4


class SugiyamaLink(LayoutInterfaceLink):
    """
    SugiyamaLink: link of the Sugiyama graph.

    Instantiated by: ../ToSugiyama.py

    :author: Nicolas Dubois
    :contact: nicdub@gmx.ch
    :version: $Revision: 1.4 $
    """
    def __init__(self, umlLink, umlFrame: UmlFrame):
        """

        Args:
            umlLink:
            umlFrame:
        """
        # ALayoutLink.__init__(self, oglObject)
        self._umlFrame: UmlFrame = umlFrame
        super().__init__(umlLink)
        self.__virtualNodes: List[VirtualSugiyamaNode] = []

    def fixControlPoints(self):
        """
        Fix a graphical path with control points.

        @author Nicolas Dubois
        """
        # Clear the actual control points of the link (not the anchor points)
        self.removeAllControlPoints()

        # Current x coordinate of the link
        x = self.getSrcAnchorPos()[0]

        # For all virtual nodes, add control points to pass through
        for virtualNode in self.__virtualNodes:
            #  ~ print "Virtual node"
            (xVNode, yVNode) = virtualNode.getPosition()
            # If link goes to up-left
            if x > xVNode:
                # Find the first real node on the right of the virtual node
                neighbor = virtualNode.getRightNode()
                #
                # Don't like embedded imports, but need to avoid cyclical dependency
                # from pyutplugins.toolplugins.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode
                from umlextensions.tools.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode

                while isinstance(neighbor, VirtualSugiyamaNode) and neighbor is not None:

                    # Try next neighbor
                    neighbor = neighbor.getRightNode()

                # If real node found
                if neighbor is not None:
                    # ctrlPoint = ControlPoint(xVNode, neighbor.getPosition()[1] + neighbor.getSize()[1])
                    ctrlPoint = UmlLineControlPoint(umlFrame=self._umlFrame,
                                                    umlLink=self._umlLink,
                                                    controlPointType=UmlLineControlPointType.LINE_POINT,
                                                    size=SUGIYAMA_CONTROL_POINT_SIZE,
                                                    x=xVNode,
                                                    y=neighbor.getPosition()[1] + neighbor.getSize()[1]
                                                    )
                    self.addControlPoint(ctrlPoint)

            else:   # If link goes to up-right
                # Don't like embedded imports, but need to avoid cyclical dependency
                # from pyutplugins.toolplugins.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode
                from umlextensions.tools.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode

                # Find the first real node on the left of the virtual node
                neighbor = virtualNode.getLeftNode()
                while isinstance(neighbor, VirtualSugiyamaNode) and neighbor is not None:

                    # Try next neighbor
                    neighbor = neighbor.getLeftNode()

                # If real node found
                if neighbor is not None:
                    # def __init__(self, x: int, y: int, parent=None):

                    # ctrlPoint = ControlPoint(xVNode, neighbor.getPosition()[1] + neighbor.getSize()[1])
                    ctrlPoint = UmlLineControlPoint(umlFrame=self._umlFrame,
                                                    umlLink=self._umlLink,
                                                    controlPointType=UmlLineControlPointType.LINE_POINT,
                                                    size=SUGIYAMA_CONTROL_POINT_SIZE,
                                                    x=xVNode,
                                                    y=neighbor.getPosition()[1] + neighbor.getSize()[1]
                                                    )
                    self.addControlPoint(ctrlPoint)

            ctrlPoint = UmlLineControlPoint(umlFrame=self._umlFrame,
                                            umlLink=self._umlLink,
                                            controlPointType=UmlLineControlPointType.LINE_POINT,
                                            size=SUGIYAMA_CONTROL_POINT_SIZE,
                                            x=xVNode,
                                            y=yVNode
                                            )
            self.addControlPoint(ctrlPoint)

    def addVirtualNode(self, node: VirtualSugiyamaNode):
        """
        Add a virtual node.

        A virtual node is inserted in long links which cross a level. If the
        link crosses more than one level, insert virtual nodes, ordered
        from source to destination (son to father - bottom-up).

        @param VirtualSugiyamaNode node : virtual node
        @author Nicolas Dubois
        """
        self.__virtualNodes.append(node)




from typing import Tuple
from typing import cast

from dataclasses import dataclass

from umlshapes.mixins.TopLeftMixin import TopLeftMixin
from umlshapes.types.UmlDimensions import UmlDimensions
from umlshapes.types.UmlPosition import UmlPosition

@dataclass
class LayoutPosition:
    x: int
    y: int


class LayoutInterfaceNode:
    """
    Interface between UML Shapes and Layout algorithms.
    """
    def __init__(self, umlShape):
        """

        Args:
            umlShape: interfaced UML Shape
        """
        self._umlShape = umlShape

    @property
    def name(self) -> str:
        """
        The class name

        Returns: name of the class
        """
        return self._umlShape.pyutClass.name

    @property
    def size(self) -> Tuple[int, int]:
        """
        Return the class size.

        Returns: (int, int): tuple (width, height)
        """
        umlDimensions: UmlDimensions = cast(TopLeftMixin, self._umlShape).size
        return umlDimensions.width, umlDimensions.height

    @property
    def position(self) -> LayoutPosition:
        """
        Get class position.

        Returns: The layout position
        """
        umlPosition: UmlPosition = cast(TopLeftMixin, self._umlShape).position
        return LayoutPosition(x=umlPosition.x, y=umlPosition.y)

    @position.setter
    def position(self, layoutPosition: LayoutPosition):
        """
        Set the class position.

        Args:
            layoutPosition
        """
        umlPosition: UmlPosition = UmlPosition(x=layoutPosition.x, y=layoutPosition.y)
        cast(TopLeftMixin, self._umlShape).position = umlPosition




from typing import cast
from typing import Dict
from typing import List
from typing import NewType

from logging import Logger
from logging import getLogger

from enum import Enum

from dataclasses import field
from dataclasses import dataclass

from wx import OK
from wx import ICON_ERROR

from wx import NewIdRef
from wx import Window
from wx import RichMessageDialog

from umlshapes.pubsubengine.UmlPubSubEngine import UmlPubSubEngine

from umlextensions.ErrorFormatter import ErrorFormatter
from umlextensions.ExtensionsPreferences import ExtensionsPreferences

from umlextensions.ExtensionsFacade import ExtensionsFacade
from umlextensions.IExtensionsFacade import IExtensionsFacade

from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName

from umlextensions.input.BaseInputExtension import BaseInputExtension
from umlextensions.input.InputPython import InputPython
from umlextensions.tools.BaseToolExtension import BaseToolExtension

from umlextensions.tools.ToolSugiyama import ToolSugiyama

from umlextensions.ExtensionsPubSub import ExtensionsPubSub


# Return type from wx.NewIdRef()
WindowId = NewType('WindowId', int)

#
#  Both of these hold the class extension types for the Extension classes
#
ExtensionList  = NewType('ExtensionList',  List[type])
ExtensionIDMap = NewType('ExtensionIDMap', Dict[WindowId, type])

def createExtensionIdMapFactory() -> ExtensionIDMap:
    return ExtensionIDMap({})

class ExtensionMapType(Enum):
    INPUT_MAP  = 'InputMap'
    OUTPUT_MAP = 'OutputMap'
    TOOL_MAP   = 'ToolMap'
    NONE       = 'None'

#
# Some nice syntactic sugar
#
@dataclass
class BaseExtensionMap:
    mapType:        ExtensionMapType = ExtensionMapType.NONE
    extensionIdMap: ExtensionIDMap   = field(default_factory=createExtensionIdMapFactory)

@dataclass
class InputExtensionMap(BaseExtensionMap):
    def __init__(self):
        super().__init__()
        self.mapType = ExtensionMapType.INPUT_MAP

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'{self.mapType} plugin count: {len(self.extensionIdMap)}'

@dataclass
class ToolExtensionMap(BaseExtensionMap):
    def __init__(self):
        super().__init__()
        self.mapType = ExtensionMapType.TOOL_MAP


TOOL_EXTENSIONS: ExtensionList = ExtensionList(
    [ToolSugiyama]
)
INPUT_EXTENSIONS: ExtensionList = ExtensionList(
    [InputPython]
)

@dataclass
class ExtensionDetails:
    name:    ExtensionName = ExtensionName('')
    author:  str = ''
    version: str = ''

class ExtensionsManager:
    """
    Manages the various extensions provided by this module

        InputExtension
        OutputExtension
        ToolExtension

    Is responsible for:

    * Identifying the extensions
    * Creating Tool, Input, & Output Menu ID References
    * Providing the callbacks that invoke the appropriate methods on the
    appropriate extensions to invoke their functionality.


    By convention prefix the plugin tool module name with the characters 'Tool'
    By convention prefix the plugin input module name with 'Input'
    By convention prefix the plugin output module name with 'Output'

    """
    def __init__(self, umlPubSubEngine: UmlPubSubEngine):

        self.logger: Logger = getLogger(__name__)

        self._extensionPreferences: ExtensionsPreferences = ExtensionsPreferences()
        self._pubsub:               ExtensionsPubSub      = ExtensionsPubSub()
        self._extensionsFacade:     IExtensionsFacade     = ExtensionsFacade(self._pubsub, umlPubSubEngine=umlPubSubEngine)
        #
        #
        #
        self._inputExtensionsMap:  InputExtensionMap = InputExtensionMap()
        self._toolExtensionsMap:   ToolExtensionMap  = ToolExtensionMap()

        # self._inputExtensionClasses:  ExtensionList = cast(ExtensionList, None)

    @property
    def extensionsPubSub(self) -> ExtensionsPubSub:
        return self._pubsub

    @property
    def inputExtensions(self) -> ExtensionList:
        """
        Get the input extension types.

        Returns:  A copy of the list of input extension classes
        """
        return ExtensionList(INPUT_EXTENSIONS[:])

    @property
    def toolExtensions(self) -> ExtensionList:
        """
        Get the tool Extensions.

        Returns:    A copy of the list of classes (the Extension classes).
        """
        return ExtensionList(TOOL_EXTENSIONS[:])

    @property
    def inputExtensionsMap(self) -> InputExtensionMap:

        if len(self._inputExtensionsMap.extensionIdMap) == 0:
            self._inputExtensionsMap.extensionIdMap = self._mapWxIdsToExtensions(self.inputExtensions)

        return self._inputExtensionsMap

    @property
    def toolExtensionsMap(self) -> ToolExtensionMap:

        if len(self._toolExtensionsMap.extensionIdMap) == 0:
            self._toolExtensionsMap.extensionIdMap = self._mapWxIdsToExtensions(TOOL_EXTENSIONS)

        return self._toolExtensionsMap

    def doImport(self, wxId: WindowId) -> ExtensionDetails:
        """
        Args:
            wxId:       The ID ref of the menu item
        """
        idMap:             ExtensionIDMap     = self.inputExtensionsMap.extensionIdMap
        clazz:             type               = idMap[wxId]
        extensionInstance: BaseInputExtension = clazz(extensionsFacade=self._extensionsFacade)

        try:
            extensionInstance.executeImport()
        except (ValueError, Exception) as e:
            self._processError(e, dialogTitle='Input Extension Error')

        return ExtensionDetails(name=extensionInstance.name, version=extensionInstance.version, author=extensionInstance.author)

    def doToolAction(self, wxId: WindowId) -> ExtensionDetails:
        """
        Args:
            wxId:   The ID ref of the menu item
        """
        idMap:        ExtensionIDMap    = self.toolExtensionsMap.extensionIdMap
        clazz:        type              = idMap[wxId]
        toolInstance: BaseToolExtension = clazz(extensionsFacade=self._extensionsFacade)

        try:
            toolInstance.executeTool()
        except (ValueError, Exception) as e:
            self._processError(e, dialogTitle='Tool Extension Error')

        return ExtensionDetails(name=toolInstance.name, version=toolInstance.version, author=toolInstance.version)

    def _processError(self, e: Exception, dialogTitle: str):
        """
        Common error handler for extensions
        Args:
            e:
            dialogTitle:
        """
        errorMessage: str = ErrorFormatter.getErrorMessage()
        fs:           str = ErrorFormatter.getErrorStack(e)
        self.logger.error(fs)
        #
        # TODO: Build my own dialog so I can set the detailed text font
        booBoo: RichMessageDialog = RichMessageDialog(cast(Window, None),
                                                      message=errorMessage,
                                                      caption=dialogTitle,
                                                      style=ICON_ERROR | OK
                                                      )
        booBoo.ShowDetailedText(fs)
        booBoo.ShowModal()

    def _mapWxIdsToExtensions(self, extensionList: ExtensionList) -> ExtensionIDMap:
        """

        Args:
            extensionList:   List of the extensions to map

        Returns:  A map of window IDs to their associated extension types
        """

        pluginMap: ExtensionIDMap = ExtensionIDMap({})

        nb: int = len(extensionList)

        for x in range(nb):
            wxId: WindowId = NewIdRef()

            pluginMap[wxId] = extensionList[x]

        return pluginMap




from typing import Union
from typing import cast
from typing import Dict
from typing import List
from typing import NewType
from typing import Callable

from logging import Logger
from logging import getLogger

from os import sep as osSep
from os import linesep as osLineSep

from antlr4 import CommonTokenStream
from antlr4 import FileStream
from antlr4.error.ErrorListener import ErrorListener

from wx import OK
from wx import ICON_ERROR
from wx import ICON_WARNING

from wx import MessageBox

from pyutmodelv2.PyutClass import PyutClass
from pyutmodelv2.PyutLink import PyutLink
from pyutmodelv2.enumerations.PyutLinkType import PyutLinkType

from umlshapes.frames.ClassDiagramFrame import ClassDiagramFrame

from umlshapes.links.eventhandlers.UmlLinkEventHandler import UmlLinkEventHandler
from umlshapes.links.eventhandlers.UmlAssociationEventHandler import UmlAssociationEventHandler

from umlshapes.pubsubengine.UmlPubSubEngine import UmlPubSubEngine

from umlshapes.shapes.eventhandlers.UmlClassEventHandler import UmlClassEventHandler

from umlshapes.shapes.UmlClass import UmlClass

from umlshapes.links.UmlAggregation import UmlAggregation
from umlshapes.links.UmlAssociation import UmlAssociation
from umlshapes.links.UmlComposition import UmlComposition
from umlshapes.links.UmlInheritance import UmlInheritance

from umlshapes.ShapeTypes import UmlAssociationGenre
from umlshapes.ShapeTypes import UmlShapes
from umlshapes.ShapeTypes import UmlLinks

from umlextensions.input.python.PythonParseException import PythonParseException

from umlextensions.input.python.pythonpegparser.PythonLexer import PythonLexer
from umlextensions.input.python.pythonpegparser.PythonParser import PythonParser

from umlextensions.input.python.visitor.ParserTypes import Associate
from umlextensions.input.python.visitor.ParserTypes import Associates
from umlextensions.input.python.visitor.ParserTypes import AssociationType
from umlextensions.input.python.visitor.ParserTypes import Associations
from umlextensions.input.python.visitor.ParserTypes import Children

from umlextensions.input.python.visitor.ParserTypes import Parents
from umlextensions.input.python.visitor.ParserTypes import ChildName
from umlextensions.input.python.visitor.ParserTypes import ParentName
from umlextensions.input.python.visitor.ParserTypes import PyutClasses
from umlextensions.input.python.visitor.ParserTypes import PyutClassName

from umlextensions.input.python.visitor.PythonPegParserClassVisitor import PythonPegParserClassVisitor
from umlextensions.input.python.visitor.PythonPegParserVisitor import PythonPegParserVisitor

ProgressCallback = Callable[[int, str], None]

UmlClassesDict = NewType('UmlClassesDict', Dict[Union[PyutClassName, ParentName, ChildName], UmlClass])


class PythonErrorListener(ErrorListener):
    #
    # Provides a default instance of {@link ConsoleErrorListener}.
    #
    # INSTANCE = None

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):

        # print("line " + str(line) + ":" + str(column) + " " + msg, file=sys.stderr)
        eMsg: str = f'{line=}{osLineSep}{column=}{osLineSep}{msg}'
        raise PythonParseException(eMsg)

class PythonToUmlShapes:
    """
    I wanted to only create 'bare' UML Classes and UML Links.  But, by virtue of how
    the OGL layer works, I have to fully prepare the shapes.  Thus, it is necessary
    for me to inject the UI frames and the pub sub engines into all the created shapes.

    """
    def __init__(self, classDiagramFrame: ClassDiagramFrame, umlPubSubEngine: UmlPubSubEngine):

        self._classDiagramFrame: ClassDiagramFrame = classDiagramFrame
        self._umlPubSubEngine:   UmlPubSubEngine   = umlPubSubEngine

        self.logger: Logger = getLogger(__name__)

        self._umlClasses: UmlShapes = UmlShapes([])
        self._umlLinks:   UmlLinks  = UmlLinks([])

        self._cumulativeParents:      Parents         = Parents({})
        """
        This dictionary is keyed by a class name that is the base class of at least one subclass.  The
        subclass value is a list of subclasses
        """
        self._cumulativeAssociations: Associations = Associations({})

    @property
    def umlLinks(self) -> UmlLinks:
        """
        Not valid until the call to .generateLinks

        Returns:
        """
        return self._umlLinks

    def pass1(self, directoryName: str, files: List[str], progressCallback: ProgressCallback) -> PyutClasses:
        """
        Uses the simplified class visitor

        Args:
            directoryName:    The fully qualified directory name of the python files
            files:            The list of files to parse
            progressCallback: A callback to report status as each file is processed; The callback
            should have the signature of the type ProgressCallback

        Returns:  The Python code parsed into model classes
        """
        pyutClasses:      PyutClasses = PyutClasses({})
        currentFileCount: int         = 0

        for fileName in files:
            try:
                fqFileName: str = f'{directoryName}{osSep}{fileName}'
                self.logger.info(f'1st pass Processing file: {fqFileName}')

                progressCallback(currentFileCount, f'Pass 1 processing: {directoryName}\n {fileName}')

                tree:    PythonParser.File_inputContext = self._setupPegBasedParser(fqFileName=fqFileName)
                visitor: PythonPegParserClassVisitor    = PythonPegParserClassVisitor()

                visitor.pyutClasses = pyutClasses
                visitor.visit(tree)
                pyutClasses = visitor.pyutClasses

            except (ValueError, Exception, PythonParseException) as e:
                if isinstance(e, PythonParseException):
                    errorMsg: str = f'{fileName}\n{e}'
                    self.logger.error(e)
                    MessageBox(errorMsg, 'Error', OK | ICON_ERROR)
                else:
                    self.logger.error(f'Error in {directoryName}/{fileName}')
                    raise e

            if len(pyutClasses) == 0:
                MessageBox('No classes processed', 'Warning', OK | ICON_WARNING)

        return pyutClasses

    def pass2(self, directoryName: str, files: List[str], pyutClasses: PyutClasses, progressCallback: Callable) -> PyutClasses:
        """
        Reverse engineering Python files to UMl Shapes

        Uses the full Python Peg Parser Visitor

        Args:
            directoryName:    The fully qualified directory name where the selected files reside
            files:            A list of files to parse
            pyutClasses:      The full list of classes scanned during pass 1
            progressCallback: The method to call to report progress
        """
        currentFileCount: int = 0
        for fileName in files:

            try:
                fqFileName: str = f'{directoryName}{osSep}{fileName}'
                self.logger.info(f'2nd pass processing file: {fqFileName}')

                progressCallback(currentFileCount, f'2nd pass - processing: {directoryName}\n {fileName}')

                tree: PythonParser.File_inputContext = self._setupPegBasedParser(fqFileName=fqFileName)
                #
                # Account for empty files
                #
                if tree is None:
                    continue

                visitor: PythonPegParserVisitor = PythonPegParserVisitor()
                #
                # Re-Initialize the visitor for this pass
                #
                visitor.pyutClasses  = pyutClasses
                visitor.parents      = self._cumulativeParents
                visitor.associations = self._cumulativeAssociations
                visitor.visit(tree)

                # Save the updated parents and associations
                self._cumulativeParents      = visitor.parents
                self._cumulativeAssociations = visitor.associations

            except (ValueError, Exception, PythonParseException) as e:
                if isinstance(e, PythonParseException):
                    errorMsg: str = f'{fileName}\n{e}'
                    self.logger.error(e)
                    MessageBox(errorMsg, 'Error', OK | ICON_ERROR)
                else:
                    raise e

        return pyutClasses

    def generateUmlClasses(self, pyutClasses: PyutClasses) -> UmlClassesDict:

        umlClassesDict: UmlClassesDict = UmlClassesDict({})
        for pyutClassName in pyutClasses:
            try:
                pyutClass: PyutClass = pyutClasses[pyutClassName]
                umlClass:  UmlClass  = UmlClass(pyutClass=pyutClass)

                eventHandler: UmlClassEventHandler = UmlClassEventHandler()
                eventHandler.SetShape(umlClass)
                eventHandler.SetPreviousHandler(umlClass.GetEventHandler())

                umlClass.SetEventHandler(eventHandler)
                umlClass.umlFrame = self._classDiagramFrame

                eventHandler.umlPubSubEngine = self._umlPubSubEngine

                umlClassesDict[PyutClassName(pyutClassName)] = umlClass

            except (ValueError, Exception) as e:
                self.logger.error(f"Error while creating class {pyutClassName},  {e}")

        return umlClassesDict

    def generateLinks(self, umlClassesDict: UmlClassesDict):
        self._generateInheritanceLinks(umlClassesDict)
        self._generateAssociationLinks(umlClassesDict)

    def _generateInheritanceLinks(self, umlClassesDict: UmlClassesDict):
        """
        Creates UML Inheritance links
        Args:
            umlClassesDict:

        """
        parents: Parents = self._cumulativeParents

        for parentName in parents.keys():
            children: Children = parents[parentName]

            for childName in children:

                try:
                    baseClass: UmlClass = umlClassesDict[parentName]
                    subClass:  UmlClass = umlClassesDict[childName]

                    umlInheritance: UmlInheritance = self._createInheritanceLink(subClass=subClass, baseClass=baseClass)

                    self._umlLinks.append(umlInheritance)
                except KeyError as ke:  # Probably there is no parent we are tracking
                    self.logger.warning(f'Apparently we are not tracking this parent:  {ke}')
                    continue

    def _generateAssociationLinks(self, umlClassesDict: UmlClassesDict):

        associations: Associations = self._cumulativeAssociations

        for className in associations:
            pyutClassName: PyutClassName = cast(PyutClassName, className)
            associates:    Associates    = associations[pyutClassName]

            for assoc in associates:
                associate: Associate = cast(Associate, assoc)
                sourceClass:      UmlClass = umlClassesDict[pyutClassName]
                destinationClass: UmlClass = umlClassesDict[associate.associateName]

                pyutLinkType: PyutLinkType   = self._toPyutLinkType(associationType=associate.associationType)
                oglLink: UmlAssociationGenre = self._createAssociationLink(sourceClass=sourceClass, destinationClass=destinationClass, linkType=pyutLinkType)

                self._umlLinks.append(oglLink)

    def _setupPegBasedParser(self, fqFileName: str) -> PythonParser.File_inputContext:
        """
        May return None if there are syntax errors in the input file
        In that case the error listener will raise and PythonParseException exception
        with the appropriate detailed error message

        Args:
            fqFileName:

        Returns:  Returns a visitor
        """

        fileStream: FileStream  = FileStream(fqFileName)
        lexer:      PythonLexer = PythonLexer(fileStream)

        stream: CommonTokenStream = CommonTokenStream(lexer)
        parser: PythonParser      = PythonParser(stream)

        parser.removeParseListeners()
        parser.addErrorListener(PythonErrorListener())

        tree: PythonParser.File_inputContext = parser.file_input()
        if parser.getNumberOfSyntaxErrors() != 0:
            eMsg: str = f"File {fqFileName} contains {parser.getNumberOfSyntaxErrors()} syntax errors"
            self.logger.error(eMsg)
            tree = cast(PythonParser.File_inputContext, None)

        return tree

    def _createInheritanceLink(self, subClass: UmlClass, baseClass: UmlClass) -> UmlInheritance:

        pyutLink: PyutLink = PyutLink("", linkType=PyutLinkType.INHERITANCE,
                                      source=subClass.pyutClass,
                                      destination=baseClass.pyutClass
                                      )

        umlInheritance: UmlInheritance = UmlInheritance(pyutLink=pyutLink, baseClass=baseClass, subClass=subClass)

        umlInheritance.umlFrame = self._classDiagramFrame
        umlInheritance.MakeLineControlPoints(n=2)       # Make this configurable

        # REMEMBER:   from subclass to base class
        subClass.addLink(umlLink=umlInheritance, destinationClass=baseClass)

        eventHandler: UmlLinkEventHandler = UmlLinkEventHandler(umlLink=umlInheritance)
        eventHandler.umlPubSubEngine = self._umlPubSubEngine
        eventHandler.SetShape(umlInheritance)
        eventHandler.SetPreviousHandler(umlInheritance.GetEventHandler())
        umlInheritance.SetEventHandler(eventHandler)

        eventHandler.umlPubSubEngine = self._umlPubSubEngine

        return umlInheritance

    def _createAssociationLink(self, sourceClass, destinationClass, linkType: PyutLinkType) -> UmlAssociationGenre:

        pyutLink: PyutLink = self._createAssociationPyutLink(linkType=linkType,
                                                             sourcePyutClass=sourceClass.pyutClass,
                                                             destinationPyutClass=destinationClass.pyutClass
                                                             )
        umlAssociation: UmlAssociationGenre

        if linkType == PyutLinkType.ASSOCIATION:
            umlAssociation = UmlAssociation(pyutLink=pyutLink)
        elif linkType == PyutLinkType.AGGREGATION:
            umlAssociation = UmlAggregation(pyutLink=pyutLink)
        elif linkType == PyutLinkType.COMPOSITION:
            umlAssociation = UmlComposition(pyutLink=pyutLink)
        else:
            assert False, 'Unknown association type'

        umlAssociation.umlFrame = self._classDiagramFrame
        umlAssociation.umlPubSubEngine = self._umlPubSubEngine
        umlAssociation.MakeLineControlPoints(n=2)       # Make this configurable

        sourceClass.addLink(umlLink=umlAssociation, destinationClass=destinationClass)

        eventHandler: UmlAssociationEventHandler = UmlAssociationEventHandler(umlAssociation=umlAssociation)
        eventHandler.umlPubSubEngine = self._umlPubSubEngine
        eventHandler.SetShape(umlAssociation)
        eventHandler.SetPreviousHandler(umlAssociation.GetEventHandler())
        umlAssociation.SetEventHandler(eventHandler)

        eventHandler.umlPubSubEngine = self._umlPubSubEngine

        return umlAssociation

    # noinspection PyUnboundLocalVariable
    def _toPyutLinkType(self, associationType: AssociationType) -> PyutLinkType:

        match associationType:
            case AssociationType.ASSOCIATION:
                pyutLinkType: PyutLinkType = PyutLinkType.ASSOCIATION
            case AssociationType.AGGREGATION:
                pyutLinkType = PyutLinkType.AGGREGATION
            case AssociationType.COMPOSITION:
                pyutLinkType = PyutLinkType.COMPOSITION
            case _:
                assert False, f'Unknown association type: {associationType.name}'

        return pyutLinkType

    def _createAssociationPyutLink(self, sourcePyutClass: PyutClass, destinationPyutClass: PyutClass, linkType: PyutLinkType) -> PyutLink:

        name: str = f' Association '
        pyutLink: PyutLink = PyutLink(name=name, source=sourcePyutClass, destination=destinationPyutClass, linkType=linkType)

        pyutLink.sourceCardinality      = 'source Cardinality'
        pyutLink.destinationCardinality = 'destination Cardinality'

        return pyutLink




from logging import Logger
from logging import getLogger

from umlshapes.ShapeTypes import UmlShapes

from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.ExtensionDataTypes import Author
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName
from umlextensions.extensiontypes.ExtensionDataTypes import Version
from umlextensions.tools.BaseToolExtension import BaseToolExtension
from umlextensions.tools.sugiyama.Sugiyama import Sugiyama


class ToolSugiyama(BaseToolExtension):

    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade=extensionsFacade)
        self.logger: Logger = getLogger(__name__)

        self._name      = ExtensionName('Sugiyama Automatic Layout')
        self._author    = Author('Nicolas Dubois <nicdub@gmx.ch>')
        self._version   = Version('1.1')

        self._autoSelectAll = False

    def setOptions(self) -> bool:
        """
        Prepare for the tool action.
        This can be used to ask some questions to the user.

        Returns: If False, the import should be cancelled.  'True' to proceed
        """
        return True

    def doAction(self):

        self.logger.info(f'Begin Sugiyama algorithm')

        selectedUmlShapes: UmlShapes = self._frameInformation.selectedUmlShapes

        sugiyama: Sugiyama = Sugiyama(extensionsFacade=self._extensionsFacade, umlFrame=self._frameInformation.umlFrame)
        sugiyama.createInterfaceOglALayout(umlShapes=selectedUmlShapes)

        sugiyama.levelFind()
        sugiyama.addVirtualNodes()
        sugiyama.barycenter()

        # noinspection PyProtectedMember
        self.logger.info(f'Number of hierarchical intersections: {sugiyama._getNbIntersectAll()}')

        sugiyama.addNonHierarchicalNodes()
        sugiyama.fixPositions()

        self._extensionsFacade.extensionModifiedProject()
        self._extensionsFacade.refreshFrame()

        self.logger.info('End Sugiyama algorithm')




from typing import Callable
from typing import NewType

from logging import Logger
from logging import getLogger

from enum import Enum

from codeallybasic.BasePubSubEngine import BasePubSubEngine
from codeallybasic.BasePubSubEngine import Topic

class ExtensionsMessageType(Enum):
    REQUEST_FRAME_INFORMATION  = 'Request Frame Information'
    REFRESH_FRAME              = 'Refresh Frame'
    EXTENSION_MODIFIED_PROJECT = 'Extension Modified Project'
    SELECT_UML_SHAPES          = 'Select UML Shapes'
    ADD_SHAPE                  = 'Add Shape'
    GET_SELECTED_UML_SHAPES    = 'Get Selected UML Shapes'
    WIGGLE_SHAPES              = 'Wiggle Shapes'


AdapterId = NewType('AdapterId', str)

EXTENSIONS_ID: AdapterId = AdapterId('sodas-teamwork-rushes-toads')


class ExtensionsPubSub(BasePubSubEngine):
    """
    Since you are using this specific pub sub engine this module uses the message
    types specifically associated with this module.  The messages
    and the UML Diagrammer needs to use this engine to subscribe to the particular message
    """
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

    # def subscribe(self, messageType: ExtensionsMessageType, adapterId: AdapterId, listener: Callable):
    def subscribe(self, messageType: ExtensionsMessageType, listener: Callable):
        self._subscribe(topic=Topic(messageType.value), listener=listener)

    # def sendMessage(self, messageType: ExtensionsMessageType, adapterId: AdapterId, **kwargs):
    def sendMessage(self, messageType: ExtensionsMessageType, **kwargs):
        self._sendMessage(topic=Topic(messageType.value), **kwargs)




from logging import Logger
from logging import getLogger

from umlshapes.ShapeTypes import UmlLinkGenre
from umlshapes.ShapeTypes import UmlShapeGenre
from umlshapes.pubsubengine.UmlPubSubEngine import UmlPubSubEngine

from umlextensions.ExtensionsTypes import FrameInformationCallback
from umlextensions.ExtensionsTypes import SelectedUmlShapesCallback

from umlextensions.IExtensionsFacade import IExtensionsFacade

from umlextensions.ExtensionsPubSub import ExtensionsMessageType
from umlextensions.ExtensionsPubSub import ExtensionsPubSub


class ExtensionsFacade(IExtensionsFacade):
    """
    This class simplifies communication between the extensions
    and the UML Diagrammer
    """

    def __init__(self, pubSub: ExtensionsPubSub, umlPubSubEngine: UmlPubSubEngine):

        self.logger: Logger = getLogger(__name__)

        self._pubsub:    ExtensionsPubSub = pubSub
        self._umlPubSub: UmlPubSubEngine  = umlPubSubEngine

    @property
    def umlPubSubEngine(self) -> UmlPubSubEngine:
        return self._umlPubSub

    def requestCurrentFrameInformation(self, callback: FrameInformationCallback):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.REQUEST_FRAME_INFORMATION, callback=callback)

    def selectUmlShapes(self):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.SELECT_UML_SHAPES)

    def getSelectedUmlShapes(self, callback: SelectedUmlShapesCallback):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.GET_SELECTED_UML_SHAPES, callback=callback)

    def extensionModifiedProject(self):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.EXTENSION_MODIFIED_PROJECT)

    def refreshFrame(self):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.REFRESH_FRAME)

    def addShape(self, umlShape: UmlShapeGenre | UmlLinkGenre):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.ADD_SHAPE, umlShape=umlShape)

    def wiggleShapes(self):
        self._pubsub.sendMessage(messageType=ExtensionsMessageType.WIGGLE_SHAPES)




from typing import List
from typing import Union

from logging import Logger
from logging import getLogger

from umlextensions.input.python.pythonpegparser.PythonParser import PythonParser

from umlextensions.input.python.visitor.ParserTypes import ChildName
from umlextensions.input.python.visitor.ParserTypes import Children
from umlextensions.input.python.visitor.ParserTypes import ParentName
from umlextensions.input.python.visitor.ParserTypes import Parents
from umlextensions.input.python.visitor.ParserTypes import PyutClassName


class ParentsDictionaryHandler:

    def __init__(self):

        self.logger: Logger = getLogger(__name__)

        self._parents:      Parents     = Parents({})

    @property
    def parents(self) -> Parents:
        return self._parents

    @parents.setter
    def parents(self, newValue: Parents):
        self._parents = newValue

    def createParentChildEntry(self, argumentsCtx: PythonParser.ArgumentsContext, childName: Union[PyutClassName, ChildName]):

        args:       PythonParser.ArgsContext = argumentsCtx.args()
        parentName: ParentName               = ParentName(args.getText())
        self.logger.debug(f'Class: {childName} is subclass of {parentName}')

        multiParents = parentName.split(',')
        if len(multiParents) > 1:
            self._handleMultiParentChild(multiParents=multiParents, childName=childName)
        else:
            self._updateParentsDictionary(parentName=parentName, childName=childName)

    def _handleMultiParentChild(self, multiParents: List[str], childName: Union[PyutClassName, ChildName]):
        """

        Args:
            multiParents:
            childName:

        """
        self.logger.debug(f'handleMultiParentChild: {childName} -- {multiParents}')
        for parent in multiParents:
            # handle the special case
            if parent.startswith('metaclass'):
                splitParent: List[str] = parent.split('=')
                parentName: ParentName = ParentName(splitParent[1])
                self._updateParentsDictionary(parentName=parentName, childName=childName)
            else:
                parentName = ParentName(parent)
                self._updateParentsDictionary(parentName=parentName, childName=childName)

    def _updateParentsDictionary(self, parentName: ParentName, childName: Union[PyutClassName, ChildName]):
        """
        Update our dictionary of parents. If the parent dictionary
        does not have an entry, create one with the single child.

        Args:
            parentName:     The prospective parent
            childName:      Child class name

        """
        if parentName in self._parents:
            children: Children = self._parents[parentName]
            children.append(childName)
        else:
            children = [childName]

        self._parents[parentName] = children




from abc import ABC
from abc import abstractmethod

from umlshapes.ShapeTypes import UmlLinkGenre
from umlshapes.ShapeTypes import UmlShapeGenre
from umlshapes.pubsubengine.UmlPubSubEngine import UmlPubSubEngine

from umlextensions.ExtensionsTypes import FrameInformation
from umlextensions.ExtensionsTypes import FrameInformationCallback
from umlextensions.ExtensionsTypes import SelectedUmlShapesCallback


class IExtensionsFacade(ABC):
    """
    This facade simplifies communication to the UML diagrammer.  This interface serves as a front-facing interface
    that masks the complexity of the UML Diagrammer
    """
    @property
    @abstractmethod
    def umlPubSubEngine(self) -> UmlPubSubEngine:
        pass

    @abstractmethod
    def requestCurrentFrameInformation(self, callback: FrameInformationCallback) -> FrameInformation:
        pass

    @abstractmethod
    def extensionModifiedProject(self):
        pass

    @abstractmethod
    def selectUmlShapes(self):
        pass

    @abstractmethod
    def getSelectedUmlShapes(self, callback: SelectedUmlShapesCallback):
        pass

    @abstractmethod
    def refreshFrame(self):
        pass

    @abstractmethod
    def addShape(self, umlShape: UmlShapeGenre | UmlLinkGenre):
        pass

    @abstractmethod
    def wiggleShapes(self):
        pass




from typing import List
from typing import Tuple
from typing import cast

from pyutmodelv2.enumerations.PyutLinkType import PyutLinkType

from umlshapes.links.UmlLink import UmlLink

from umlshapes.shapes.UmlLineControlPoint import UmlLineControlPoint

from umlextensions.tools.sugiyama.RealSugiyamaNode import RealSugiyamaNode
from umlextensions.tools.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode

SyntheticNode = RealSugiyamaNode | VirtualSugiyamaNode


class LayoutInterfaceLink:
    """
    Interface between Uml Links and Layout algorithms.

    Layout algorithms use this interface to access the links on the
    diagram.

    * The interface protects the structure of the diagram.
    * The diagram structure and diagram m methods can be changed.

    Thus, we only need to update is this interface, not the automatic layout algorithm.

    Note:
        This happened in 2025, when Humberto did away with the miniogl home grown code and
        instead used wxPython's wx.lib.ogl code
    """
    def __init__(self, umlLink: UmlLink):
        """

        Args:
            umlLink:
        """
        self._umlLink: UmlLink = umlLink
        self._srcNode: SyntheticNode = cast(SyntheticNode, None)
        self._dstNode: SyntheticNode = cast(SyntheticNode, None)

    @property
    def linkType(self) -> PyutLinkType:
        """
        Return the link type

        Returns: Link type
        """
        return self._umlLink.pyutLink.linkType

    @property
    def source(self) -> SyntheticNode:
        """

        Returns:  the source node.
        """
        return self._srcNode

    @source.setter
    def source(self, node: SyntheticNode):
        """
        Set the source node.

        Args:
            node:  The link source node
        """
        self._srcNode = node

    @property
    def destination(self) -> SyntheticNode:
        """

        Returns: The destination node
        """
        return self._dstNode

    @destination.setter
    def destination(self, node: SyntheticNode):
        """
        Set the destination node.

        Args:
            node: The new link destination
        """
        self._dstNode = node

    def setSrcAnchorPos(self, x: int, y: int):
        """
        Set anchor position (absolute coordinates) on source class.

        Args:
            x:
            y:
        """
        umLink: UmlLink = self._umlLink
        x1, y1, x2, y2 = umLink.GetEnds()

        umLink.SetEnds(x1=x, y1=y, x2=x2, y2=y2)
        # self._umlLink.sourceAnchor.SetPosition(x, y)

    def setDestAnchorPos(self, x: int, y: int):
        """
        Set anchor position (absolute coordinates) on destination class.

        Args:
            x:
            y:
        """
        umLink: UmlLink = self._umlLink
        x1, y1, x2, y2 = umLink.GetEnds()

        umLink.SetEnds(x1=x1, y1=y1, x2=x, y2=y)

        # self._umlLink.destinationAnchor.SetPosition(x, y)

    def getSrcAnchorPos(self):
        """
        Get anchor position (absolute coordinates) on source class.

        Returns:    (int, int) : tuple with (x, y) coordinates
        """
        umLink: UmlLink = self._umlLink
        x1, y1, x2, y2 = umLink.GetEnds()

        return x1, y1
        # return self._umlLink.sourceAnchor.GetPosition()

    def getDestAnchorPos(self):
        """
        Return anchor position (absolute coordinates) on destination class.

        Returns:  (int, int) : tuple with (x, y) coordinates
        """
        umLink: UmlLink = self._umlLink
        x1, y1, x2, y2 = umLink.GetEnds()

        return x2, y2

    # noinspection PyUnusedLocal
    def addControlPoint(self, control: UmlLineControlPoint, last=None):
        """
        Add a control point. If the parameter last present, add a point right after last.
        TODO: the 'last' parameter is used by no one

        Args:
            control:  control point to add
            last:     add control right after last
        """
        # self._oglLink.AddControl(control, last)
        pt: Tuple[int, int] = control.position.x, control.position.y
        self._umlLink.InsertLineControlPoint(point=pt)

    def removeControlPoint(self, controlPoint):
        """
        Remove a control point.

        TODO:  This is not used by the Sugiyama algorithm

        Args:
            controlPoint: control point to remove
        """
        umlLink: UmlLink = self._umlLink

        controlPoints: List[UmlLineControlPoint] = umlLink.GetLineControlPoints()
        controlPoints.remove(controlPoint)

        # self._umlLink.Remove(controlPoint)

    def removeAllControlPoints(self):
        """
        Remove all control points.
        """
        # self._umlLink.ResetControlPoints()
        self._umlLink.DeleteControlPoints()
        # self._umlLink.RemoveAllControlPoints()




from typing import cast

from logging import Logger
from logging import getLogger

from abc import ABC
from abc import abstractmethod

from wx import BeginBusyCursor
from wx import EndBusyCursor

from wx import Yield as wxYield

from umlextensions.ExtensionsTypes import FrameInformation
from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.BaseExtension import BaseExtension


class BaseToolExtension(BaseExtension, ABC):
    """
    Base class for extensions that can manipulate UML diagrams.  Examples,
    include but are not limited to:

        * Various layouts (Sugiyama, Orthogonal, Force Directed
        * Arranging non crossing links
        * Arranging orthogonal links
    """
    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade)
        self.logger: Logger = getLogger(__name__)

        self._frameInformation: FrameInformation = cast(FrameInformation, None)

    def executeTool(self):
        """
        This is used by the Plugin Manger to invoke the tool.  This should NOT
        be overridden
        """
        if self.setOptions() is True:

            if self._autoSelectAll is True:
                self._extensionsFacade.selectUmlShapes()
            self._extensionsFacade.requestCurrentFrameInformation(callback=self._executeTool)

    def _executeTool(self, frameInformation: FrameInformation):

        self._frameInformation = frameInformation

        if frameInformation.frameActive is False:
            self.showNoUmlFrameDialog()
        else:
            self._selectedOglObjects = frameInformation.selectedUmlShapes  # syntactic sugar

            if len(self._selectedOglObjects) == 0 and self._requireSelection is True:
                self.showNoSelectedUmlShapesDialog()
            else:
                BeginBusyCursor()
                wxYield()
                self.doAction()
                EndBusyCursor()

    @abstractmethod
    def setOptions(self) -> bool:
        """
        Prepare for the tool action
        This can be used to query the user for additional plugin options

        Returns: If False, cancel the tool action 'True' to proceed
        """
        pass

    @abstractmethod
    def doAction(self):
        """
        Do the tool's action
        """
        pass



from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from antlr4 import ParserRuleContext
from antlr4.tree.Tree import TerminalNodeImpl
from codeallybasic.ConfigurationProperties import PropertyName
from pyutmodelv2.PyutClass import PyutClass
from pyutmodelv2.PyutField import PyutField
from pyutmodelv2.PyutType import PyutType
from pyutmodelv2.enumerations.PyutVisibility import PyutVisibility

from umlextensions.input.python.pythonpegparser.PythonParser import PythonParser
from umlextensions.input.python.visitor.ParserTypes import PyutClassName
from umlextensions.input.python.visitor.ParserTypes import PyutClasses

from umlextensions.input.python.pythonpegparser.PythonParserVisitor import PythonParserVisitor


NO_CLASS_DEF_CONTEXT: PythonParser.Class_defContext = cast(PythonParser.Class_defContext, None)

class BaseVisitor(PythonParserVisitor):
    def __init__(self):
        super().__init__()
        self.baseLogger: Logger = getLogger(__name__)

        self._pyutClasses:  PyutClasses = PyutClasses({})

    def _isThisAssignmentInsideAMethod(self, ctx: PythonParser.AssignmentContext) -> bool:

        ans: bool = False

        currentCtx: ParserRuleContext = self._extractMethodContext(ctx=ctx)
        if currentCtx is not None:
            ans = True

        return ans

    def _makeFieldForClass(self, className: PyutClassName, propertyName: Union[PropertyName, str], typeStr: str, defaultValue: str):
        """

        Args:
            className:
            propertyName:
            typeStr:
            defaultValue:
        """
        pyutField: PyutField = PyutField(name=propertyName, type=PyutType(typeStr), visibility=PyutVisibility.PUBLIC, defaultValue=defaultValue)
        pyutClass: PyutClass = self._pyutClasses[className]

        pyutClass.fields.append(pyutField)

    def _extractClassName(self, ctx: PythonParser.Class_defContext) -> PyutClassName:
        """
        Get a class name from a Class_defContext
        Args:
            ctx:

        Returns:    A class name
        """

        child:     PythonParser.Class_def_rawContext = ctx.class_def_raw()
        # name:      TerminalNodeImpl                  = child.NAME()
        name:      TerminalNodeImpl = child.name()
        className: PyutClassName    = name.getText()

        return className

    def _extractClassDefContext(self, ctx: ParserRuleContext) -> PythonParser.Class_defContext:
        """
        Args:
            ctx:

        Returns:  Either a class definition context or the sentinel value NO_CLASS_DEF_CONTEXT
        """
        currentCtx: ParserRuleContext = ctx
        while currentCtx.parentCtx:
            if isinstance(currentCtx, PythonParser.Class_defContext):
                return currentCtx
            currentCtx = currentCtx.parentCtx

        return NO_CLASS_DEF_CONTEXT

    def _extractMethodContext(self, ctx: ParserRuleContext) -> PythonParser.Function_defContext:

        currentCtx: ParserRuleContext = ctx

        while isinstance(currentCtx, PythonParser.Function_defContext) is False:
            currentCtx = currentCtx.parentCtx
            if currentCtx is None:
                break

        if currentCtx is not None:
            raw: PythonParser.Function_def_rawContext = cast(PythonParser.Function_defContext, currentCtx).function_def_raw()
            self.baseLogger.debug(f'Found method: {raw.name()}')

        return cast(PythonParser.Function_defContext, currentCtx)

    def _findArgListContext(self, ctx: PythonParser.Class_defContext) -> PythonParser.ArgumentsContext:

        argumentsCtx: PythonParser.ArgumentsContext = cast(PythonParser.ArgumentsContext, None)

        classDefRawContext: PythonParser.Class_def_rawContext = ctx.class_def_raw()
        for childCtx in classDefRawContext.children:
            if isinstance(childCtx, PythonParser.ArgumentsContext):
                argumentsCtx = childCtx
                break

        return argumentsCtx




from typing import List

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from os import linesep as osLineSep
from os import sep as osSep


@dataclass
class CodeLine:
    fileName:   str = ''
    lineNumber: str = ''
    methodName: str = ''
    code:       str = ''


CodeLines       = List[CodeLine]
CompressedLines = List[str]
StackTraceList  = List[str]


class StackTraceFormatter:
    """
    Get a stack trace list as follows:
        ```python
        import traceback
            trString: str = traceback.extract_stack()

            fmtString: StackTraceList = traceback.format_list(trString)
        ```
    """
    SPACE: str = ' '
    COMMA: str = ','
    METHOD_SPLITTER: str = ' in'

    FILE_NAME_RETRIEVAL_LENGTH: int = 24

    def __init__(self, stackTraceList: StackTraceList, codeFileNameLength: int = FILE_NAME_RETRIEVAL_LENGTH):

        self.logger: Logger = getLogger(__name__)

        self._stackTraceList:      StackTraceList  = stackTraceList
        self._codeLines:           CodeLines       = []
        self._compressedCodeLines: CompressedLines = []
        self._codeFileNameLength:  int             = codeFileNameLength

    @property
    def codeLines(self) -> CodeLines:
        if len(self._codeLines) == 0:
            self._parseStack()

        return self._codeLines

    @property
    def compressedCodeLines(self) -> List[str]:

        if len(self._codeLines) == 0:
            self._parseStack()
        if len(self._compressedCodeLines) == 0:
            self._compressTheStack()

        return self._compressedCodeLines

    def dumpedStackList(self) -> str:
        if len(self._codeLines) == 0:
            self._parseStack()
        if len(self._compressedCodeLines) == 0:
            self._compressTheStack()

        bigString: str = ''
        for compressedLine in self._compressedCodeLines:
            bigString = f'{bigString}{osLineSep}{compressedLine}'

        return bigString

    def _parseStack(self):

        for rawCodeLine in self._stackTraceList:
            codeLine: CodeLine = self._parseRawCodeLine(rawCodeLine)
            self._codeLines.append(codeLine)

    def _compressTheStack(self):
        """
        Assumes the stack has been parsed
        """
        for codeLine in self._codeLines:
            compressedLine = (
                f'{codeLine.fileName:24} '
                f'{codeLine.lineNumber:5}'
                f'{codeLine.methodName:16} '
                f'{codeLine.code:24}'
            )
            self._compressedCodeLines.append(compressedLine)

    def _parseRawCodeLine(self, rawCodeLine: str) -> CodeLine:

        codeLine: CodeLine  = CodeLine()
        lfParse:  List[str] = rawCodeLine.split(osLineSep)

        fileLineNumMethod: str = lfParse[0]
        sourceCode:  str = lfParse[1]

        codeLine.code       = sourceCode.strip(StackTraceFormatter.SPACE)
        codeLine.lineNumber = self._getLineNumber(fileLineNum=fileLineNumMethod)
        codeLine.methodName = self._getMethodName(fileLineNumMethod=fileLineNumMethod)
        codeLine.fileName   = self._getFileName(fileLineNum=fileLineNumMethod)

        return codeLine

    def _getLineNumber(self, fileLineNum: str) -> str:
        """

        Args:
            fileLineNum: The part of the raw code line that contains the file name and the line number

        Returns:  The text of the line number
        """
        flnParse: List[str] = fileLineNum.split(StackTraceFormatter.COMMA)

        rawLineNumberParse: str = flnParse[1]

        tmpList: List[str] = rawLineNumberParse.strip(StackTraceFormatter.SPACE).split(StackTraceFormatter.SPACE)
        lineNumber: str = tmpList[1]

        return lineNumber

    def _getMethodName(self, fileLineNumMethod: str) -> str:

        flnParse: List[str] = fileLineNumMethod.split(StackTraceFormatter.COMMA)

        rawMethodParse: str       = flnParse[2]
        methodParse:    List[str] = rawMethodParse.split(StackTraceFormatter.METHOD_SPLITTER)
        fullMethod:     str       = methodParse[1]

        return fullMethod

    def _getFileName(self, fileLineNum: str) -> str:
        """
        Does not necessarily return an entire code path;  Retrieves a default number of characters;  Can
        be controlled by the class consumer

        Args:
            fileLineNum: The part of the raw code line that contains the filename and the line number

        Returns:  The text of the filename
        """
        flnParse:         List[str] = fileLineNum.split(StackTraceFormatter.COMMA)
        rawFileNameParse: str       = flnParse[0]

        truncatedFileName: str = rawFileNameParse[-StackTraceFormatter.FILE_NAME_RETRIEVAL_LENGTH:]
        modifiedFileName:  str = self.__stripUpToFirstDirectorySeparator(longFileName=truncatedFileName)

        return modifiedFileName

    def __stripUpToFirstDirectorySeparator(self, longFileName: str):

        if longFileName.startswith(osSep):
            modifiedFileName: str = longFileName
        else:
            osSepIdx: int = longFileName.find(osSep)
            if osSepIdx == -1:
                modifiedFileName = longFileName
            else:
                modifiedFileName = longFileName[osSepIdx:]

        return modifiedFileName

    def __str__(self) -> str:
        return self.__repr_()

    def __repr_(self) -> str:

        retStr:  str = ''
        lineNum: int = 0
        for line in self._stackTraceList:
            lineNum += 1
            retStr = f'{retStr}\n{lineNum} - {line}'

        return retStr



# Generated from PythonParser.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .PythonParser import PythonParser
else:
    from PythonParser import PythonParser

# This class defines a complete generic visitor for a parse tree produced by PythonParser.

class PythonParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by PythonParser#file_input.
    def visitFile_input(self, ctx:PythonParser.File_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#interactive.
    def visitInteractive(self, ctx:PythonParser.InteractiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#eval.
    def visitEval(self, ctx:PythonParser.EvalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#func_type.
    def visitFunc_type(self, ctx:PythonParser.Func_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#statements.
    def visitStatements(self, ctx:PythonParser.StatementsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#statement.
    def visitStatement(self, ctx:PythonParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#statement_newline.
    def visitStatement_newline(self, ctx:PythonParser.Statement_newlineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#simple_stmts.
    def visitSimple_stmts(self, ctx:PythonParser.Simple_stmtsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#simple_stmt.
    def visitSimple_stmt(self, ctx:PythonParser.Simple_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#compound_stmt.
    def visitCompound_stmt(self, ctx:PythonParser.Compound_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#assignment.
    def visitAssignment(self, ctx:PythonParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#annotated_rhs.
    def visitAnnotated_rhs(self, ctx:PythonParser.Annotated_rhsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#augassign.
    def visitAugassign(self, ctx:PythonParser.AugassignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#return_stmt.
    def visitReturn_stmt(self, ctx:PythonParser.Return_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#raise_stmt.
    def visitRaise_stmt(self, ctx:PythonParser.Raise_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#global_stmt.
    def visitGlobal_stmt(self, ctx:PythonParser.Global_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#nonlocal_stmt.
    def visitNonlocal_stmt(self, ctx:PythonParser.Nonlocal_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#del_stmt.
    def visitDel_stmt(self, ctx:PythonParser.Del_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#yield_stmt.
    def visitYield_stmt(self, ctx:PythonParser.Yield_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#assert_stmt.
    def visitAssert_stmt(self, ctx:PythonParser.Assert_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#import_stmt.
    def visitImport_stmt(self, ctx:PythonParser.Import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#import_name.
    def visitImport_name(self, ctx:PythonParser.Import_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#import_from.
    def visitImport_from(self, ctx:PythonParser.Import_fromContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#import_from_targets.
    def visitImport_from_targets(self, ctx:PythonParser.Import_from_targetsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#import_from_as_names.
    def visitImport_from_as_names(self, ctx:PythonParser.Import_from_as_namesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#import_from_as_name.
    def visitImport_from_as_name(self, ctx:PythonParser.Import_from_as_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#dotted_as_names.
    def visitDotted_as_names(self, ctx:PythonParser.Dotted_as_namesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#dotted_as_name.
    def visitDotted_as_name(self, ctx:PythonParser.Dotted_as_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#dotted_name.
    def visitDotted_name(self, ctx:PythonParser.Dotted_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#block.
    def visitBlock(self, ctx:PythonParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#decorators.
    def visitDecorators(self, ctx:PythonParser.DecoratorsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#class_def.
    def visitClass_def(self, ctx:PythonParser.Class_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#class_def_raw.
    def visitClass_def_raw(self, ctx:PythonParser.Class_def_rawContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#function_def.
    def visitFunction_def(self, ctx:PythonParser.Function_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#function_def_raw.
    def visitFunction_def_raw(self, ctx:PythonParser.Function_def_rawContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#params.
    def visitParams(self, ctx:PythonParser.ParamsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#parameters.
    def visitParameters(self, ctx:PythonParser.ParametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#slash_no_default.
    def visitSlash_no_default(self, ctx:PythonParser.Slash_no_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#slash_with_default.
    def visitSlash_with_default(self, ctx:PythonParser.Slash_with_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_etc.
    def visitStar_etc(self, ctx:PythonParser.Star_etcContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#kwds.
    def visitKwds(self, ctx:PythonParser.KwdsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#param_no_default.
    def visitParam_no_default(self, ctx:PythonParser.Param_no_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#param_no_default_star_annotation.
    def visitParam_no_default_star_annotation(self, ctx:PythonParser.Param_no_default_star_annotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#param_with_default.
    def visitParam_with_default(self, ctx:PythonParser.Param_with_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#param_maybe_default.
    def visitParam_maybe_default(self, ctx:PythonParser.Param_maybe_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#param.
    def visitParam(self, ctx:PythonParser.ParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#param_star_annotation.
    def visitParam_star_annotation(self, ctx:PythonParser.Param_star_annotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#annotation.
    def visitAnnotation(self, ctx:PythonParser.AnnotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_annotation.
    def visitStar_annotation(self, ctx:PythonParser.Star_annotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#default_assignment.
    def visitDefault_assignment(self, ctx:PythonParser.Default_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#if_stmt.
    def visitIf_stmt(self, ctx:PythonParser.If_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#elif_stmt.
    def visitElif_stmt(self, ctx:PythonParser.Elif_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#else_block.
    def visitElse_block(self, ctx:PythonParser.Else_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#while_stmt.
    def visitWhile_stmt(self, ctx:PythonParser.While_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#for_stmt.
    def visitFor_stmt(self, ctx:PythonParser.For_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#with_stmt.
    def visitWith_stmt(self, ctx:PythonParser.With_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#with_item.
    def visitWith_item(self, ctx:PythonParser.With_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#try_stmt.
    def visitTry_stmt(self, ctx:PythonParser.Try_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#except_block.
    def visitExcept_block(self, ctx:PythonParser.Except_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#except_star_block.
    def visitExcept_star_block(self, ctx:PythonParser.Except_star_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#finally_block.
    def visitFinally_block(self, ctx:PythonParser.Finally_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#match_stmt.
    def visitMatch_stmt(self, ctx:PythonParser.Match_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#subject_expr.
    def visitSubject_expr(self, ctx:PythonParser.Subject_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#case_block.
    def visitCase_block(self, ctx:PythonParser.Case_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#guard.
    def visitGuard(self, ctx:PythonParser.GuardContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#patterns.
    def visitPatterns(self, ctx:PythonParser.PatternsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#pattern.
    def visitPattern(self, ctx:PythonParser.PatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#as_pattern.
    def visitAs_pattern(self, ctx:PythonParser.As_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#or_pattern.
    def visitOr_pattern(self, ctx:PythonParser.Or_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#closed_pattern.
    def visitClosed_pattern(self, ctx:PythonParser.Closed_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#literal_pattern.
    def visitLiteral_pattern(self, ctx:PythonParser.Literal_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#literal_expr.
    def visitLiteral_expr(self, ctx:PythonParser.Literal_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#complex_number.
    def visitComplex_number(self, ctx:PythonParser.Complex_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#signed_number.
    def visitSigned_number(self, ctx:PythonParser.Signed_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#signed_real_number.
    def visitSigned_real_number(self, ctx:PythonParser.Signed_real_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#real_number.
    def visitReal_number(self, ctx:PythonParser.Real_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#imaginary_number.
    def visitImaginary_number(self, ctx:PythonParser.Imaginary_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#capture_pattern.
    def visitCapture_pattern(self, ctx:PythonParser.Capture_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#pattern_capture_target.
    def visitPattern_capture_target(self, ctx:PythonParser.Pattern_capture_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#wildcard_pattern.
    def visitWildcard_pattern(self, ctx:PythonParser.Wildcard_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#value_pattern.
    def visitValue_pattern(self, ctx:PythonParser.Value_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#attr.
    def visitAttr(self, ctx:PythonParser.AttrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#name_or_attr.
    def visitName_or_attr(self, ctx:PythonParser.Name_or_attrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#group_pattern.
    def visitGroup_pattern(self, ctx:PythonParser.Group_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#sequence_pattern.
    def visitSequence_pattern(self, ctx:PythonParser.Sequence_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#open_sequence_pattern.
    def visitOpen_sequence_pattern(self, ctx:PythonParser.Open_sequence_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#maybe_sequence_pattern.
    def visitMaybe_sequence_pattern(self, ctx:PythonParser.Maybe_sequence_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#maybe_star_pattern.
    def visitMaybe_star_pattern(self, ctx:PythonParser.Maybe_star_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_pattern.
    def visitStar_pattern(self, ctx:PythonParser.Star_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#mapping_pattern.
    def visitMapping_pattern(self, ctx:PythonParser.Mapping_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#items_pattern.
    def visitItems_pattern(self, ctx:PythonParser.Items_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#key_value_pattern.
    def visitKey_value_pattern(self, ctx:PythonParser.Key_value_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#double_star_pattern.
    def visitDouble_star_pattern(self, ctx:PythonParser.Double_star_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#class_pattern.
    def visitClass_pattern(self, ctx:PythonParser.Class_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#positional_patterns.
    def visitPositional_patterns(self, ctx:PythonParser.Positional_patternsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#keyword_patterns.
    def visitKeyword_patterns(self, ctx:PythonParser.Keyword_patternsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#keyword_pattern.
    def visitKeyword_pattern(self, ctx:PythonParser.Keyword_patternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_alias.
    def visitType_alias(self, ctx:PythonParser.Type_aliasContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_params.
    def visitType_params(self, ctx:PythonParser.Type_paramsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_param_seq.
    def visitType_param_seq(self, ctx:PythonParser.Type_param_seqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_param.
    def visitType_param(self, ctx:PythonParser.Type_paramContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_param_bound.
    def visitType_param_bound(self, ctx:PythonParser.Type_param_boundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_param_default.
    def visitType_param_default(self, ctx:PythonParser.Type_param_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_param_starred_default.
    def visitType_param_starred_default(self, ctx:PythonParser.Type_param_starred_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#expressions.
    def visitExpressions(self, ctx:PythonParser.ExpressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#expression.
    def visitExpression(self, ctx:PythonParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#yield_expr.
    def visitYield_expr(self, ctx:PythonParser.Yield_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_expressions.
    def visitStar_expressions(self, ctx:PythonParser.Star_expressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_expression.
    def visitStar_expression(self, ctx:PythonParser.Star_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_named_expressions.
    def visitStar_named_expressions(self, ctx:PythonParser.Star_named_expressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_named_expression.
    def visitStar_named_expression(self, ctx:PythonParser.Star_named_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#assignment_expression.
    def visitAssignment_expression(self, ctx:PythonParser.Assignment_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#named_expression.
    def visitNamed_expression(self, ctx:PythonParser.Named_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#disjunction.
    def visitDisjunction(self, ctx:PythonParser.DisjunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#conjunction.
    def visitConjunction(self, ctx:PythonParser.ConjunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#inversion.
    def visitInversion(self, ctx:PythonParser.InversionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#comparison.
    def visitComparison(self, ctx:PythonParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#compare_op_bitwise_or_pair.
    def visitCompare_op_bitwise_or_pair(self, ctx:PythonParser.Compare_op_bitwise_or_pairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#eq_bitwise_or.
    def visitEq_bitwise_or(self, ctx:PythonParser.Eq_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#noteq_bitwise_or.
    def visitNoteq_bitwise_or(self, ctx:PythonParser.Noteq_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lte_bitwise_or.
    def visitLte_bitwise_or(self, ctx:PythonParser.Lte_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lt_bitwise_or.
    def visitLt_bitwise_or(self, ctx:PythonParser.Lt_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#gte_bitwise_or.
    def visitGte_bitwise_or(self, ctx:PythonParser.Gte_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#gt_bitwise_or.
    def visitGt_bitwise_or(self, ctx:PythonParser.Gt_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#notin_bitwise_or.
    def visitNotin_bitwise_or(self, ctx:PythonParser.Notin_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#in_bitwise_or.
    def visitIn_bitwise_or(self, ctx:PythonParser.In_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#isnot_bitwise_or.
    def visitIsnot_bitwise_or(self, ctx:PythonParser.Isnot_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#is_bitwise_or.
    def visitIs_bitwise_or(self, ctx:PythonParser.Is_bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#bitwise_or.
    def visitBitwise_or(self, ctx:PythonParser.Bitwise_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#bitwise_xor.
    def visitBitwise_xor(self, ctx:PythonParser.Bitwise_xorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#bitwise_and.
    def visitBitwise_and(self, ctx:PythonParser.Bitwise_andContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#shift_expr.
    def visitShift_expr(self, ctx:PythonParser.Shift_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#sum.
    def visitSum(self, ctx:PythonParser.SumContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#term.
    def visitTerm(self, ctx:PythonParser.TermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#factor.
    def visitFactor(self, ctx:PythonParser.FactorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#power.
    def visitPower(self, ctx:PythonParser.PowerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#await_primary.
    def visitAwait_primary(self, ctx:PythonParser.Await_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#primary.
    def visitPrimary(self, ctx:PythonParser.PrimaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#slices.
    def visitSlices(self, ctx:PythonParser.SlicesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#slice.
    def visitSlice(self, ctx:PythonParser.SliceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#atom.
    def visitAtom(self, ctx:PythonParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#group.
    def visitGroup(self, ctx:PythonParser.GroupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambdef.
    def visitLambdef(self, ctx:PythonParser.LambdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_params.
    def visitLambda_params(self, ctx:PythonParser.Lambda_paramsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_parameters.
    def visitLambda_parameters(self, ctx:PythonParser.Lambda_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_slash_no_default.
    def visitLambda_slash_no_default(self, ctx:PythonParser.Lambda_slash_no_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_slash_with_default.
    def visitLambda_slash_with_default(self, ctx:PythonParser.Lambda_slash_with_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_star_etc.
    def visitLambda_star_etc(self, ctx:PythonParser.Lambda_star_etcContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_kwds.
    def visitLambda_kwds(self, ctx:PythonParser.Lambda_kwdsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_param_no_default.
    def visitLambda_param_no_default(self, ctx:PythonParser.Lambda_param_no_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_param_with_default.
    def visitLambda_param_with_default(self, ctx:PythonParser.Lambda_param_with_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_param_maybe_default.
    def visitLambda_param_maybe_default(self, ctx:PythonParser.Lambda_param_maybe_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#lambda_param.
    def visitLambda_param(self, ctx:PythonParser.Lambda_paramContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#fstring_middle.
    def visitFstring_middle(self, ctx:PythonParser.Fstring_middleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#fstring_replacement_field.
    def visitFstring_replacement_field(self, ctx:PythonParser.Fstring_replacement_fieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#fstring_conversion.
    def visitFstring_conversion(self, ctx:PythonParser.Fstring_conversionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#fstring_full_format_spec.
    def visitFstring_full_format_spec(self, ctx:PythonParser.Fstring_full_format_specContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#fstring_format_spec.
    def visitFstring_format_spec(self, ctx:PythonParser.Fstring_format_specContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#fstring.
    def visitFstring(self, ctx:PythonParser.FstringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#string.
    def visitString(self, ctx:PythonParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#strings.
    def visitStrings(self, ctx:PythonParser.StringsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#list.
    def visitList(self, ctx:PythonParser.ListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#tuple.
    def visitTuple(self, ctx:PythonParser.TupleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#set.
    def visitSet(self, ctx:PythonParser.SetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#dict.
    def visitDict(self, ctx:PythonParser.DictContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#double_starred_kvpairs.
    def visitDouble_starred_kvpairs(self, ctx:PythonParser.Double_starred_kvpairsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#double_starred_kvpair.
    def visitDouble_starred_kvpair(self, ctx:PythonParser.Double_starred_kvpairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#kvpair.
    def visitKvpair(self, ctx:PythonParser.KvpairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#for_if_clauses.
    def visitFor_if_clauses(self, ctx:PythonParser.For_if_clausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#for_if_clause.
    def visitFor_if_clause(self, ctx:PythonParser.For_if_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#listcomp.
    def visitListcomp(self, ctx:PythonParser.ListcompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#setcomp.
    def visitSetcomp(self, ctx:PythonParser.SetcompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#genexp.
    def visitGenexp(self, ctx:PythonParser.GenexpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#dictcomp.
    def visitDictcomp(self, ctx:PythonParser.DictcompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#arguments.
    def visitArguments(self, ctx:PythonParser.ArgumentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#args.
    def visitArgs(self, ctx:PythonParser.ArgsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#kwargs.
    def visitKwargs(self, ctx:PythonParser.KwargsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#starred_expression.
    def visitStarred_expression(self, ctx:PythonParser.Starred_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#kwarg_or_starred.
    def visitKwarg_or_starred(self, ctx:PythonParser.Kwarg_or_starredContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#kwarg_or_double_starred.
    def visitKwarg_or_double_starred(self, ctx:PythonParser.Kwarg_or_double_starredContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_targets.
    def visitStar_targets(self, ctx:PythonParser.Star_targetsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_targets_list_seq.
    def visitStar_targets_list_seq(self, ctx:PythonParser.Star_targets_list_seqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_targets_tuple_seq.
    def visitStar_targets_tuple_seq(self, ctx:PythonParser.Star_targets_tuple_seqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_target.
    def visitStar_target(self, ctx:PythonParser.Star_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#target_with_star_atom.
    def visitTarget_with_star_atom(self, ctx:PythonParser.Target_with_star_atomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#star_atom.
    def visitStar_atom(self, ctx:PythonParser.Star_atomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#single_target.
    def visitSingle_target(self, ctx:PythonParser.Single_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#single_subscript_attribute_target.
    def visitSingle_subscript_attribute_target(self, ctx:PythonParser.Single_subscript_attribute_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#t_primary.
    def visitT_primary(self, ctx:PythonParser.T_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#del_targets.
    def visitDel_targets(self, ctx:PythonParser.Del_targetsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#del_target.
    def visitDel_target(self, ctx:PythonParser.Del_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#del_t_atom.
    def visitDel_t_atom(self, ctx:PythonParser.Del_t_atomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#type_expressions.
    def visitType_expressions(self, ctx:PythonParser.Type_expressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#func_type_comment.
    def visitFunc_type_comment(self, ctx:PythonParser.Func_type_commentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#name_except_underscore.
    def visitName_except_underscore(self, ctx:PythonParser.Name_except_underscoreContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PythonParser#name.
    def visitName(self, ctx:PythonParser.NameContext):
        return self.visitChildren(ctx)



del PythonParser



from typing import List
from typing import NewType
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import field
from dataclasses import dataclass

from wx import DEFAULT_DIALOG_STYLE
from wx import FD_CHANGE_DIR
from wx import FD_FILE_MUST_EXIST
from wx import FD_MULTIPLE
from wx import FD_OPEN
from wx import FileDialog
from wx import RESIZE_BORDER
from wx import STAY_ON_TOP
from wx import ID_CANCEL
from wx import ID_OK
from wx import OK
from wx import CANCEL

from wx import EVT_BUTTON

from wx import Button
from wx import CommandEvent

from wx.grid import Grid

from wx.lib.sized_controls import SizedDialog
from wx.lib.sized_controls import SizedPanel

from umlextensions.input.InputFormat import InputFormat

ImportModules = NewType('ImportModules', List[str])


def importModulesFactory() -> ImportModules:
    return ImportModules([])


@dataclass
class Package:
    packageName:    str = ''
    importModules:  ImportModules = field(default_factory=importModulesFactory)


ImportPackages = NewType('ImportPackages', List[Package])


class DlgSelectMultiplePackages(SizedDialog):
    """
    TODO:  This might be useful outside of importing Python files.
    """

    def __init__(self, startDirectory: str, inputFormat: InputFormat):

        style: int = RESIZE_BORDER | STAY_ON_TOP | DEFAULT_DIALOG_STYLE
        super().__init__(None, title='Select Multiple Modules', style=style)

        self.logger: Logger = getLogger(__name__)

        self._startDirectory: str         = startDirectory
        self._inputFormat:    InputFormat = inputFormat

        sizedPanel:           SizedPanel = self.GetContentsPane()
        sizedPanel.SetSizerType('vertical')
        sizedPanel.SetSizerProps(expand=True, proportion=1)

        self._btnMore:    Button = cast(Button, None)
        self._btnCancel:  Button = cast(Button, None)
        self._btnOk:      Button = cast(Button, None)
        self._simpleGrid: Grid   = cast(Grid, None)

        self._layoutSimpleGrid(parent=sizedPanel)
        self._layoutCustomDialogButtonContainer(parent=sizedPanel)

        self._importPackages: ImportPackages = ImportPackages([])
        self._packageCount:   int            = 0
        self._moduleCount:    int            = 0

        self._currentGridRow: int = 0

        self._resizeDialog()

    def _layoutSimpleGrid(self, parent: SizedPanel):

        simpleGrid: Grid = Grid(parent)
        simpleGrid.CreateGrid(numRows=1, numCols=2)

        simpleGrid.SetColLabelValue(0, 'Package Name')
        simpleGrid.SetColLabelValue(1, 'Module Count')

        simpleGrid.AutoSizeColumns()

        self._simpleGrid = simpleGrid

    def _layoutCustomDialogButtonContainer(self, parent: SizedPanel, ):
        """
        Create Ok and Cancel
        Since we want to use a custom button set, we will not use the
        CreateStdDialogBtnSizer here, we'll just create our own panel with
        a horizontal layout and add the buttons to that

        Args:
            parent:
        """
        buttonPanel: SizedPanel = SizedPanel(parent)
        buttonPanel.SetSizerType('horizontal')
        buttonPanel.SetSizerProps(expand=False, halign='right')  # expand False allows aligning right

        #
        # Layout custom buttons here
        #
        self._btnMore   = Button(buttonPanel, label='&More')
        self._btnCancel = Button(buttonPanel, ID_CANCEL, '&Cancel')
        self._btnOk     = Button(buttonPanel, ID_OK, '&Ok')

        self.Bind(EVT_BUTTON, self._onMore,  self._btnMore)
        self.Bind(EVT_BUTTON, self._onOk,    self._btnOk)
        self.Bind(EVT_BUTTON, self._onClose, self._btnCancel)

        self._btnOk.SetDefault()

    @property
    def importPackages(self) -> ImportPackages:
        """
        Only valid if user pressed 'ok'

        Returns: the selected import directories
        """
        return self._importPackages

    @property
    def packageCount(self) -> int:
        return self._packageCount

    @property
    def moduleCount(self) -> int:
        return self._moduleCount

    # noinspection PyUnusedLocal
    def _onMore(self, event: CommandEvent):

        style: int = FD_OPEN | FD_FILE_MUST_EXIST | FD_MULTIPLE | FD_CHANGE_DIR

        with FileDialog(None, "Choose files to import", wildcard=self._composeWildCardSpecification(), defaultDir=self._startDirectory, style=style) as dlg:
            if dlg.ShowModal() == ID_OK:
                importDirectory: Package = Package()

                importDirectory.packageName   = dlg.GetDirectory()
                importDirectory.importModules = ImportModules(dlg.GetFilenames())

                self._packageCount += 1
                currentModuleCount: int = len(importDirectory.importModules)
                self._moduleCount  += currentModuleCount
                self._importPackages.append(importDirectory)
                self._simpleGrid.SetCellValue(self._currentGridRow, 0, importDirectory.packageName)
                self._simpleGrid.SetCellValue(self._currentGridRow, 1, str(currentModuleCount))

                self._simpleGrid.AppendRows(1)
                self._currentGridRow += 1
                self._simpleGrid.AutoSizeColumns()
                self._resizeDialog()
            else:
                self._importPackages = ImportPackages([])

    # noinspection PyUnusedLocal
    def _onOk(self, event: CommandEvent):
        """
        """
        self.EndModal(OK)

    # noinspection PyUnusedLocal
    def _onClose(self, event: CommandEvent):
        """
        """
        self.EndModal(CANCEL)

    def _composeWildCardSpecification(self) -> str:

        inputFormat: InputFormat = self._inputFormat

        wildcard: str = (
            f'{inputFormat.formatName} '
            f' (*, {inputFormat.fileSuffix}) '
            f'|*.{inputFormat.fileSuffix}'
        )
        return wildcard

    def _resizeDialog(self):
        """
        A little trick to make sure that you can't resize the dialog to
        less screen space than the controls need
        """
        self.Fit()
        self.SetMinSize(self.GetSize())




from umlextensions.extensiontypes.BaseFormat import BaseFormat

from umlextensions.extensiontypes.ExtensionDataTypes import FileSuffix
from umlextensions.extensiontypes.ExtensionDataTypes import FormatName
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionDescription


class InputFormat(BaseFormat):
    """
    Syntactic sugar
    """
    def __init__(self, formatName: FormatName, fileSuffix: FileSuffix, description: ExtensionDescription):
        super().__init__(formatName=formatName, fileSuffix=fileSuffix, description=description)

